var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
export function FormControlMixin(SuperClass) {
    var _FormControl_instances, _FormControl_focused, _FormControl_forceError, _FormControl_touched, _FormControl_formValidationGroup_get, _FormControl_value, _FormControl_onFocus, _FormControl_onBlur, _FormControl_onInvalid, _FormControl_shouldShowError, _FormControl_validate;
    class FormControl extends SuperClass {
        /* eslint-disable  @typescript-eslint/no-explicit-any */
        constructor(...args) {
            super(...args);
            _FormControl_instances.add(this);
            /** The ElementInternals instance for the control. */
            this.internals = this.attachInternals();
            /**
             * Keep track of if the control has focus
             * @private
             */
            _FormControl_focused.set(this, false);
            /**
             * Exists to control when an error should be displayed
             * @private
             */
            _FormControl_forceError.set(this, false);
            /**
             * Toggles to true whenever the element has been focused. This property
             * will reset whenever the control's formResetCallback is called.
             * @private
             */
            _FormControl_touched.set(this, false);
            /**
             * Acts as a cache for the current value so the value can be re-evaluated
             * whenever an attribute changes or on some other event.
             */
            _FormControl_value.set(this, '');
            /**
             * Set this[touched] and this[focused]
             * to true when the element is focused
             * @private
             */
            _FormControl_onFocus.set(this, () => {
                __classPrivateFieldSet(this, _FormControl_touched, true, "f");
                __classPrivateFieldSet(this, _FormControl_focused, true, "f");
                __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
            });
            /**
             * Reset this[focused] on blur
             * @private
             */
            _FormControl_onBlur.set(this, () => {
                __classPrivateFieldSet(this, _FormControl_focused, false, "f");
                __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_validate).call(this, this.shouldFormValueUpdate() ? __classPrivateFieldGet(this, _FormControl_value, "f") : '');
                /**
                 * Set forceError to ensure error messages persist until
                 * the value is changed.
                 */
                if (!this.validity.valid && __classPrivateFieldGet(this, _FormControl_touched, "f")) {
                    __classPrivateFieldSet(this, _FormControl_forceError, true, "f");
                }
                const showError = __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
                if (this.validationMessageCallback) {
                    this.validationMessageCallback(showError ? this.internals.validationMessage : '');
                }
            });
            /**
             * For the show error state on invalid
             * @private
             */
            _FormControl_onInvalid.set(this, () => {
                var _a;
                __classPrivateFieldSet(this, _FormControl_forceError, true, "f");
                __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
                (_a = this === null || this === void 0 ? void 0 : this.validationMessageCallback) === null || _a === void 0 ? void 0 : _a.call(this, this.showError ? this.internals.validationMessage : '');
            });
            this.addEventListener('focus', __classPrivateFieldGet(this, _FormControl_onFocus, "f"));
            this.addEventListener('blur', __classPrivateFieldGet(this, _FormControl_onBlur, "f"));
            this.addEventListener('invalid', __classPrivateFieldGet(this, _FormControl_onInvalid, "f"));
            this.setValue(null);
        }
        /** Wires up control instances to be form associated */
        static get formAssociated() {
            return true;
        }
        static get validators() {
            return this.formControlValidators || [];
        }
        /**
         * Allows the FormControl instance to respond to Validator attributes.
         * For instance, if a given Validator has a `required` attribute, that
         * validator will be evaluated whenever the host's required attribute
         * is updated.
         */
        static get observedAttributes() {
            const validatorAttributes = this.validators.map((validator) => validator.attribute);
            const observedAttributes = super.observedAttributes || [];
            /** Make sure there are no duplicates inside the attributes list */
            const attributeSet = new Set([...observedAttributes, ...validatorAttributes]);
            return [...attributeSet];
        }
        /**
         * Return the validator associated with a given attribute. If no
         * Validator is associated with the attribute, it will return null.
         */
        static getValidator(attribute) {
            return this.validators.find((validator) => validator.attribute === attribute) || null;
        }
        /** Return a reference to the control's form */
        get form() {
            return this.internals.form;
        }
        /**
         * Will return true if it is recommended that the control shows an internal
         * error. If using this property, it is wise to listen for 'invalid' events
         * on the element host and call preventDefault on the event. Doing this will
         * prevent browsers from showing a validation popup.
         */
        get showError() {
            return __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
        }
        /**
         * Forward the internals checkValidity method
         * will return the valid state of the control.
         */
        checkValidity() {
            return this.internals.checkValidity();
        }
        /** The element's validity state */
        get validity() {
            return this.internals.validity;
        }
        /**
         * The validation message shown by a given Validator object. If the control
         * is in a valid state this should be falsy.
         */
        get validationMessage() {
            return this.internals.validationMessage;
        }
        attributeChangedCallback(name, oldValue, newValue) {
            var _a;
            (_a = super.attributeChangedCallback) === null || _a === void 0 ? void 0 : _a.call(this, name, oldValue, newValue);
            /**
             * Check to see if a Validator is associated with the changed attribute.
             * If one exists, call control's validate function which will perform
             * control validation.
             */
            const proto = this.constructor;
            const validator = proto.getValidator(name);
            if (validator && this.validationTarget) {
                this.setValue(__classPrivateFieldGet(this, _FormControl_value, "f"));
            }
        }
        /** PUBLIC LIFECYCLE METHODS */
        /**
         * Sets the control's form value if the call to `shouldFormValueUpdate`
         * returns `true`.
         * @param value {FormValue} - The value to pass to the form
         */
        setValue(value) {
            var _a;
            __classPrivateFieldSet(this, _FormControl_forceError, false, "f");
            (_a = this.validationMessageCallback) === null || _a === void 0 ? void 0 : _a.call(this, '');
            __classPrivateFieldSet(this, _FormControl_value, value, "f");
            const valueShouldUpdate = this.shouldFormValueUpdate();
            const valueToUpdate = valueShouldUpdate ? value : null;
            this.internals.setFormValue(valueToUpdate);
            __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_validate).call(this, valueToUpdate);
            __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
        }
        /**
         * This method can be overridden to determine if the control's form value
         * should be set on a call to `setValue`. An example of when a user might want
         * to skip this step is when implementing checkbox-like behavior, first checking
         * to see if `this.checked` is set to a truthy value. By default this returns
         * `true`.
         */
        shouldFormValueUpdate() {
            return true;
        }
        /** Reset control state when the form is reset */
        formResetCallback() {
            var _a, _b;
            __classPrivateFieldSet(this, _FormControl_touched, false, "f");
            __classPrivateFieldSet(this, _FormControl_forceError, false, "f");
            __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this);
            (_a = this.resetFormControl) === null || _a === void 0 ? void 0 : _a.call(this);
            (_b = this.validationMessageCallback) === null || _b === void 0 ? void 0 : _b.call(this, __classPrivateFieldGet(this, _FormControl_instances, "m", _FormControl_shouldShowError).call(this) ? this.validationMessage : '');
        }
    }
    _FormControl_focused = new WeakMap(), _FormControl_forceError = new WeakMap(), _FormControl_touched = new WeakMap(), _FormControl_value = new WeakMap(), _FormControl_onFocus = new WeakMap(), _FormControl_onBlur = new WeakMap(), _FormControl_onInvalid = new WeakMap(), _FormControl_instances = new WeakSet(), _FormControl_formValidationGroup_get = function _FormControl_formValidationGroup_get() {
        const rootNode = this.getRootNode();
        const selector = `${this.localName}[name="${this.getAttribute('name')}"]`;
        return rootNode.querySelectorAll(selector);
    }, _FormControl_shouldShowError = function _FormControl_shouldShowError() {
        if (this.hasAttribute('disabled')) {
            return false;
        }
        const showError = __classPrivateFieldGet(this, _FormControl_forceError, "f") || (__classPrivateFieldGet(this, _FormControl_touched, "f") && !this.validity.valid && !__classPrivateFieldGet(this, _FormControl_focused, "f"));
        /**
         * At the time of writing Firefox doesn't support states
         * TODO: Remove when check for states when fully support is in place
         */
        if (showError && this.internals.states) {
            this.internals.states.add('--show-error');
        }
        else if (this.internals.states) {
            this.internals.states.delete('--show-error');
        }
        return showError;
    }, _FormControl_validate = function _FormControl_validate(value) {
        const proto = this.constructor;
        const validity = {};
        let validationMessage = '';
        let controlIsValid = true;
        proto.validators.forEach((validator) => {
            /** A validator should only be evaluated if that key isn't currently in an invalid state */
            if (validity[validator.key || 'customError'] !== true) {
                /** Get data oof the Validator */
                const { message, isValid } = validator;
                /** If a key is not set, use `customError` as a catch-all */
                const key = validator.key || 'customError';
                /** Invoke the Validator isValid callback with the instance and the value */
                const valid = isValid(this, value);
                /**
                 * Invert the validity because we are setting the new property
                 * on the new ValidityState object
                 */
                validity[key] = !valid;
                if (valid === false && validationMessage === '') {
                    controlIsValid = false;
                    let messageResult = '';
                    /**
                    * The Validator interfaces allows for the message property
                    * to be either a string or a function. If it is a function,
                    * we want to get the returned value to use when calling
                    * ElementInternals.prototype.setValidity.
                    *
                    * If the Validator.message is a string, use it directly. However,
                    * if a control has a ValidityCallback, it can override the error
                    * message for a given validity key.
                    */
                    if (this.validityCallback && this.validityCallback(key)) {
                        messageResult = this.validityCallback(key);
                    }
                    else if (message instanceof Function) {
                        messageResult = message(this, value);
                    }
                    else if (typeof message === 'string') {
                        messageResult = message;
                    }
                    validationMessage = messageResult;
                }
            }
        });
        /**
         * In some cases, the validationTarget might not be rendered
         * at this point, if the validationTarget does exist, proceed
         * with a call to internals.setValidity. If the validationTarget
         * is still not set, we essentially wait a tick until it is there.
         *
         * If the validityTarget does not exist even after the setTimeout,
         * this will throw.
         */
        if (controlIsValid) {
            this.internals.setValidity({});
            /** If the element is part of a formControlValidationGroup, reset those values */
            if (proto.formControlValidationGroup === true) {
                __classPrivateFieldGet(this, _FormControl_instances, "a", _FormControl_formValidationGroup_get).forEach(control => {
                    /** Don't duplicate effort */
                    if (control !== this) {
                        control.internals.setValidity({});
                    }
                });
            }
        }
        else if (this.validationTarget) {
            this.internals.setValidity(validity, validationMessage, this.validationTarget);
        }
        else {
            /**
             * If the validationTarget is not set, the user can decide how they would
             * prefer to handle focus when the field is validated.
             */
            this.internals.setValidity(validity, validationMessage);
            /**
             * It could be that a give component hasn't rendered by the time it is first
             * validated. If it hasn't been, wait a bit and add the validationTarget
             * to the setValidity call.
             *
             * TODO: Document the edge case that an element doesn't have a validationTarget
             * and must be focusable some other way
             */
            let tick = 0;
            const id = setInterval(() => {
                if (tick >= 100 || this.validity.valid) {
                    clearInterval(id);
                }
                else if (this.validationTarget) {
                    this.internals.setValidity(this.validity, validationMessage, this.validationTarget);
                    clearInterval(id);
                }
                tick += 1;
            }, 0);
        }
    };
    return FormControl;
}
//# sourceMappingURL=FormControlMixin.js.map